% This functoin simulates the performance of LUPA with a  DRL setup
% but in the Simulink environment
% Written by Shangyan 7/29/2024
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clc
clear

%% General Parameters
T_f = 600;  % consistent with simulation final time
T_s = 0.5;  % agent sample time (note which is different compared
% to simulation time step which is directly specified in Simulink) 
dt = 0.01; % s, it is noted that this is the simulation step

%% Configurate WEC-Sim
% Initialize WEC-Sim
SimFileName = 'Sim_DRL_LUPA_ver5_NonlinearSWTrainS1';
Hs = 0.21;
Tp = 3.09;
Ki_max = 0;
Ki_min = -6e3;
Kp_max = 3000;
Kp_min = 200;
N_w = 500;
run('wecSimInputFile');
clear simu waves body cable pto constraint ptoSim mooring 
runWecSimCML = 1;
run('initializeWecSim');

%% Configurate Environment Interface
% Action Info spec
% all possible actions
ActionSpac{1} = [100;0];
ActionSpac{2} = [0;50];
ActionSpac{3} = [0;0];
ActionSpac{4} = [-100;0];
ActionSpac{5} = [0;-50];
ActInfo = rlFiniteSetSpec(ActionSpac);
ActInfo.Name = 'Del K';
% Observation Info spec
ObsInfo = rlNumericSpec([2 1]);
ObsInfo.Name = 'State';
ObsInfo.Description = 'z,v';
% create the environment object
env = rlSimulinkEnv(SimFileName,strcat(SimFileName,'/DRL Control/DQN Agent'),ObsInfo,ActInfo);
% create the reset function
env.ResetFcn = @(in) RL_Reset(in);
env.UseFastRestart = "Off";

%% Create DQN Agent
% Define number of inputs and outpus of the DNN
NoI = ObsInfo.Dimension(1);   % number of inputs
NoO = numel(ActInfo.Elements);  % number of outputs
NL = 256;   % number of neurons in the hidden layers
% create the neural network
criticNet = [
    featureInputLayer(NoI,'Normalization','none','Name','state')
    fullyConnectedLayer(NL,'Name','fc1')
    reluLayer('Name','relu1')
    fullyConnectedLayer(NL,'Name','fc2')
    reluLayer('Name','relu2')
    fullyConnectedLayer(NoO,'Name','fc3')];
% Display number of parameters to be learned
criticNet = dlnetwork(criticNet);
summary(criticNet)
% Define the critics
critic = rlVectorQValueFunction(criticNet,ObsInfo,ActInfo);
criticOptions = rlOptimizerOptions('LearnRate',3e-4,'GradientThreshold',1,'L2RegularizationFactor',1e-4);
% Define the Agent options
% UseDoubleDQN: enable this to use double network to improve convergence
% ExperienceBuffer:  How many experience will be stored as a memorial pool
% MiniBatchSize: How many random data will be extracted from memorial pool
agentOptions = rlDQNAgentOptions(...
    'SampleTime',T_s,...
    'CriticOptimizerOptions',criticOptions,...
    'UseDoubleDQN',true,...
    'ExperienceBufferLength',1e4,...
    'MiniBatchSize',256,...
    'TargetSmoothFactor',1e-3,...
    'DiscountFactor',0.99);
% agentOptions.EpsilonGreedyExploration.Epsilon = 1e-7;
agentOptions.EpsilonGreedyExploration.EpsilonDecay = 5e-3;
agentOptions.EpsilonGreedyExploration.EpsilonMin = 0.01;
% Finalize the definition of the DQN agent
% It is noted that the output name shall be consistent with the agent name
% in the Simulink model
agent = rlDQNAgent(critic,agentOptions);

%% Train agent
% Specify the training options
% rng(0);
trainingOptions = rlTrainingOptions(...
    'MaxEpisodes',1200,...
    'MaxStepsPerEpisode',ceil(T_f/T_s),...
    'Verbose',false,...
    'Plots','training-progress',...
    'StopTrainingCriteria','AverageReward',...
    'StopTrainingValue',5e8,...
    'SaveAgentCriteria','EpisodeReward',...
    'SaveAgentValue',4e3,....
    'SaveAgentDirectory', pwd + "\run\Agents",...
    'SimulationStorageType','none');
% Save the DRL setup for postprocessing
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% train
trainingStats = train(agent,env,trainingOptions);
save Results_DQN























